module SmartDExp where

import Word
import Bits hiding (rotate)
import Maybe
import Monad  ( mplus )
import IOExts

import Types
import DExp
import PPDExp
import PP
import qualified F32vec4 as V4
import Int

infixl 8  <<<, >>>  -- left, right shift
infix  4  ==*, /=*, <*, <=*, >=*, >*
infixr 3  &&*
infixr 2  ||*
infixl 1  >>=*
infix  4  ==%

-- ToDo: make it work with GHC (a la Cse.hs)
-- primitive ptrEq' "unsafePtrEq" :: a -> a -> Bool

--
-- A cheap test for Exp equality might be useful for
-- some Exp trees.
--
(==%) :: Exp -> Exp -> Bool
x ==% y = {-unsafePtrEq x y ||-} x == y

-- Sometimes it's useful to suppress optimization.  In that case, define
-- "optimize" as False.  We'll still do constant folding.  (Should there
-- be a separate flag for constant folding?)
optimize = True

-- Whether to float conditionals upward.  See also okFloatIf below for finer control
floatIfs = True


zero :: Type -> Exp

intZero     = LitInt     0
doubleZero  = LitDouble  0
f32vec4Zero = PackF32 doubleZero doubleZero doubleZero doubleZero

zero Int     = intZero
zero Double  = doubleZero
zero F32vec4 = f32vec4Zero 
zero ty      = error ("zero: undefined for type " ++ show ty)


one :: Type -> Exp

intOne     = LitInt     1
doubleOne  = LitDouble  1
f32vec4One = PackF32 doubleOne doubleOne doubleOne doubleOne

one Int     = intOne
one Double  = doubleOne
one F32vec4 = f32vec4One
one ty      = error ("one: undefined for type " ++ show ty)


-- Supported "literal" types: Int, Bool, and Double

class Literal a where
  mkLit :: a -> Exp
  dsLit :: Exp -> Maybe a

instance Literal Int where
  mkLit n          = LitInt n
  dsLit (LitInt n) = Just n
  dsLit _          = Nothing

instance Literal Bool where
  mkLit n           = LitBool n
  dsLit (LitBool n) = Just n
  dsLit _           = Nothing

instance Literal Double where
  mkLit n             = LitDouble n
  dsLit (LitDouble n) = Just n
  dsLit _             = Nothing

instance Literal V4.F32vec4 where
  -- This one is different.  Use the PackF32 constructor applied to
  -- literals.  Simplifies the code somewhat.  Also reflects that to the
  -- code generator literals are not special, since they cost to
  -- construct.
  -- Recall that memory and constructor order are opposite.
  mkLit (a,b,c,d) = PackF32 (LitDouble d) (LitDouble c) (LitDouble b) (LitDouble a)
  dsLit (PackF32 (LitDouble d) (LitDouble c) (LitDouble b) (LitDouble a)) = Just (a,b,c,d)
  dsLit _              = Nothing
  -- Old version:
  -- mkLit n              = LitF32vec4 n
  -- dsLit (LitF32vec4 n) = Just n


-- Generalized constant folders

class CFold1 f where
  cfold1 :: f -> Exp -> Maybe Exp

class CFold2 f where
  cfold2 :: f -> Exp -> Exp -> Maybe Exp

class CFold3 f where
  cfold3 :: f -> Exp -> Exp -> Exp -> Maybe Exp

class CFold4 f where
  cfold4 :: f -> Exp -> Exp -> Exp -> Exp -> Maybe Exp



-- Trivial, always-failing constant folder.
data NoFold = NoFold

instance CFold1 NoFold where cfold1 _ _       = Nothing
instance CFold2 NoFold where cfold2 _ _ _     = Nothing
instance CFold3 NoFold where cfold3 _ _ _ _   = Nothing
instance CFold4 NoFold where cfold4 _ _ _ _ _ = Nothing


-- Literal->literal functions are constant folders

instance (Literal a, Literal b) => CFold1 (a -> b) where
  cfold1 f e1 = do a <- dsLit e1
                   return (mkLit (f a))

instance (Literal a, Literal b, Literal c) => CFold2 (a -> b -> c) where
  cfold2 f e1 e2 = do a <- dsLit e1
                      b <- dsLit e2
                      return (mkLit (f a b))

instance (Literal a, Literal b, Literal c, Literal d) =>
     CFold3 (a -> b -> c -> d) where
  cfold3 f e1 e2 e3 = do a <- dsLit e1
                         b <- dsLit e2
                         c <- dsLit e3
                         return (mkLit (f a b c))

instance (Literal a, Literal b, Literal c, Literal d, Literal e) =>
     CFold4 (a -> b -> c -> d -> e) where
  cfold4 f e1 e2 e3 e4 = do a <- dsLit e1
                            b <- dsLit e2
                            c <- dsLit e3
                            d <- dsLit e4
                            return (mkLit (f a b c d))




-- We can also pair up constant folders to make a constant folder that
-- tries the first and on failure the second.  Useful for overloaded
-- functions.  (See "==" and "+" below.)

instance (CFold1 f, CFold1 g) => CFold1 (f,g) where
  cfold1 (f,g) e1 = cfold1 f e1 `mplus` cfold1 g e1

instance (CFold2 f, CFold2 g) => CFold2 (f,g) where
  cfold2 (f,g) e1 e2 = cfold2 f e1 e2 `mplus` cfold2 g e1 e2

instance (CFold3 f, CFold3 g) => CFold3 (f,g) where
  cfold3 (f,g) e1 e2 e3 = cfold3 f e1 e2 e3 `mplus` cfold3 g e1 e2 e3

instance (CFold4 f, CFold4 g) => CFold4 (f,g) where
  cfold4 (f,g) e1 e2 e3 e4 =
    cfold4 f e1 e2 e3 e4 `mplus` cfold4 g e1 e2 e3 e4


-- Optimizers:  Try if-floating, then constant folding, then
-- function-specific optimization.

opt1 :: CFold1 f => (Exp -> Exp) -> f -> (Exp -> Exp)
opt1 opt f = h
 where
  h (If c a b) | okFloatIf [] = ifE c (h a) (h b)
  h e1         = maybe (opt e1) id (cfold1 f e1)

{-
 Don't float ifs - needed to break a loop between the overloaded 'opt1'
 and the non-overloaded 'notE' (this loop is only a problem with GHC, as
 it insists that all type sigs in a group have the same contexts, something
 that's definitely not the case for 'notE' and 'opt1'.)
-}

opt1_noIf :: CFold1 f => (Exp -> Exp) -> f -> (Exp -> Exp)

opt1_noIf opt f = h
 where
  h e1 = maybe (opt e1) id (cfold1 f e1)

opt2 :: CFold2 f => (Exp -> Exp -> Exp)
     -> f -> (Exp -> Exp -> Exp)
opt2 opt f = h
 where
  h (If c a b) e2 | okFloatIf [e2] = ifE c (h a e2) (h b e2)
  h e1 (If c a b) | okFloatIf [e1] = ifE c (h e1 a) (h e1 b)
  h e1 e2 = maybe (opt e1 e2) id (cfold2 f e1 e2)

opt3 opt f = h
 where
  h (If c a b) e2 e3 | okFloatIf [e2,e3] = ifE c (h a e2 e3) (h b e2 e3)
  h e1 (If c a b) e3 | okFloatIf [e1,e3] = ifE c (h e1 a e3) (h e1 b e3)
  h e1 e2 (If c a b) | okFloatIf [e1,e2] = ifE c (h e1 e2 a) (h e1 e2 b)
  h e1 e2 e3 = maybe (opt e1 e2 e3) id (cfold3 f e1 e2 e3)

opt4 opt f = h
 where
  h (If c a b) e2 e3 e4 | okFloatIf [e2,e3,e4] = ifE c (h a e2 e3 e4) (h b e2 e3 e4)
  h e1 (If c a b) e3 e4 | okFloatIf [e1,e3,e4] = ifE c (h e1 a e3 e4) (h e1 b e3 e4)
  h e1 e2 (If c a b) e4 | okFloatIf [e1,e2,e4] = ifE c (h e1 e2 a e4) (h e1 e2 b e4)
  h e1 e2 e3 (If c a b) | okFloatIf [e1,e2,e3] = ifE c (h e1 e2 e3 a) (h e1 e2 e3 b)
  h e1 e2 e3 e4 = maybe (opt e1 e2 e3 e4) id (cfold4 f e1 e2 e3 e4)

-- Whether it's okay to float ifs, given the expressions to be replicated
okFloatIf :: [Exp] -> Bool

okFloatIf es = floatIfs &&
  True  -- always float
  --all isSimple es  -- nothing complex is replicated.

-- Look for smarter definitions for okFloatIf.


-- Make an optN function that either uses the given optimizer or the given
-- constructor.
mbOpt optN | optimize  = \ opt f con -> optN opt f
           | otherwise = \ opt f con -> optN con f

opt1C :: CFold1 f => (Exp -> Exp) -> f -> (Exp -> Exp) -> (Exp -> Exp)
opt1C = mbOpt opt1

opt1_noIfC :: CFold1 f => (Exp -> Exp) -> f -> (Exp -> Exp) -> (Exp -> Exp)
opt1_noIfC = mbOpt opt1_noIf

opt2C :: CFold2 f => (Exp -> Exp -> Exp) -> f -> (Exp -> Exp -> Exp) -> (Exp -> Exp -> Exp)
opt2C = mbOpt opt2

opt3C :: CFold3 f => (Exp -> Exp -> Exp -> Exp) -> f
      -> (Exp -> Exp -> Exp -> Exp) -> (Exp -> Exp -> Exp -> Exp)
opt3C = mbOpt opt3

opt4C :: CFold4 f => (Exp -> Exp -> Exp -> Exp -> Exp) -> f
      -> (Exp -> Exp -> Exp -> Exp -> Exp) -> (Exp -> Exp -> Exp -> Exp -> Exp)
opt4C = mbOpt opt4




------- Conditional

ifZ :: Exp -> Exp -> Exp -> Exp

-- Merge nested conditionals that have the same condition.  The
-- combination of "if splitting" done by the (a,b) and Color Syntactic
-- instances, and the if lifting done by opt<n>, yields many mergeable
-- conditionals.
ifZ c (If c' a' b') b | c ==% c' = ifE c a' b
ifZ c a (If c' a' b') | c ==% c' = ifE c a b'
ifZ _ a b             | a ==% b  = a
ifZ c a b                        = If c (assume c a) (assume (notE c) b)

-- The first two clauses appear to be redundant with the last.  ToDo:
-- remove them and check that we lose no reductions.

{-

-- Could use this version instead of the previous one.  This
-- implementation collects all conditionals, but one is probably
-- sufficient, since simplification is done bottom-up.
ifZ c a b                       = assume [] (If c a b)


-- Simplify a sub-expression given a list of boolean expressions known to
-- be true.  Currently, only contained conditionals are simplified.

assume :: [Exp] -> Exp -> Exp
assume asms (If c a b) 
  |      c `elem` asms = assume asms a
  | notE c `elem` asms = assume asms b
  | otherwise          = If c (assume (     c : asms) a)
                              (assume (notE c : asms) b)

-- Could recursively simplify non-if sub-expressions here.  If so, memoize.
assume assume e = e
-}

-- This implementation uses just a single condition.
assume :: Exp -> Exp -> Exp
assume asm = assumeRec
 where
   assumeRec (If c a b) 
    |      c ==% asm  = assumeRec a
    | notE c ==% asm  = assumeRec b
    | otherwise       = If c (assumeRec a) (assumeRec b)

   assumeRec e = e


-- The general function, doing various general optimizations.
ifE :: Exp -> Exp -> Exp -> Exp

ifE (LitBool True ) a b = a
ifE (LitBool False) a b = b
ifE (Not c')        a b = ifE c' b a
ifE (If c d e)      a b = ifE c (ifE d a b) (ifE e a b)
--ifE c a b | a == b      = a
ifE c a b               = ifZ c a b

-- Is the if/if rule a good idea?  It certainly bloats.  I did it because
-- a and b often have c-based conditionals as well.  Moving them under the
-- "if c" gives them a chance to simplify.  Experimental.


------ Type-specific simplifiers

--- Boolean

instance Ord Exp where
  min = opt2C minZ ((min :: Binop Int       , min :: Binop Double)
                  ,(min :: Binop V4.F32vec4, min :: Binop Bool)) Min
        where
          minZ e e' | e ==% e'  = e
                    | otherwise = Min e e'
  max = opt2C maxZ ((max :: Binop Int, max :: Binop Double)
                  ,(max :: Binop V4.F32vec4, max :: Binop Bool)) Max
        where
          maxZ e e' | e ==% e'  = e
                    | otherwise = Max e e'
  compare = error "Exp: no compare method.  Use <*, etc."

-- Comparisons
type Cmp a = a -> a -> Bool

true  = LitBool True
false = LitBool False

(==*), (/=*) :: Exp -> Exp -> Exp

(==*) = opt2C (==#) (((==) :: Cmp Int, (==) :: Cmp Double), (==) :: Cmp Bool) Equal
 where
   e ==# e' | e ==% e' = true
   -- Commute literals to RHS
   e@(LitDouble _) ==# e'  =  e' ==* e
   e@(LitInt _)    ==# e'  =  e' ==* e
   -- e@(LitF32vec4 _) ==# e'  =  e' ==* e
   -- ToDo: maybe add rules for boolean literals
   -- c+e==c' or c*e==c' for literals c,c'
   -- (LitDouble x `Add` e) ==# LitDouble x' = e ==* LitDouble (x' - x)
   e ==# (e' `Add` e'') | isLit e  && isLit e'  = e - e'  ==* e''
                        | isLit e  && isLit e'' = e - e'' ==* e'
   (LitDouble x `Mul` e) ==# LitDouble x' = e ==* LitDouble (x' / x)
   I2f e ==# LitDouble x | r == 0    = e ==* LitInt k
                         | otherwise = false
     where (k,r) = properFraction x
   e ==# e' | negTest e e' = -e ==* -e'
   e ==# e'           = Equal e e'

(<*), (<=*), (>=*), (>*) :: Exp -> Exp -> Exp

(<*) = opt2C (<#) ((<) :: Cmp Int, (<) :: Cmp Double) LessThan
 where
   -- Cannot commute literals to RHS this time, since "<" is not
   -- commutative, and ">" would just gets commuted back right away.
   -- c+e<c' or c*e<c' for literals c,c'
   -- (LitDouble x `Add` e) <# LitDouble x' = e <* LitDouble (x' - x)
   -- LitDouble x <# (LitDouble x' `Add` e') = LitDouble (x - x') <* e'
   (e `Add` e') <# e'' | isLit e  && isLit e'' = e' <* e'' - e
                       | isLit e' && isLit e'' = e  <* e'' - e'
   e <# (e' `Add` e'') | isLit e  && isLit e'  = e - e'  <* e''
                       | isLit e  && isLit e'' = e - e'' <* e'
   (e `Mul` LitDouble x) <# LitDouble x' =
     (if x > 0 then (<*) else (>*)) e (LitDouble (x' / x))
   LitDouble x <# (e' `Mul` LitDouble x') =
     (if x' > 0 then (<*) else (>*)) (LitDouble (x / x')) e'
   I2f e <# LitDouble x = e <* LitInt (ceiling x)
   LitDouble x <# I2f e = LitInt (floor x) <* e
   e <# e' | negTest e e' = -e' <* -e

   e <# e' =  LessThan e e'

-- Both args are negative or one is and the other is a literal.  We can
-- then simplify comparators.
negTest :: Exp -> Exp -> Bool
negTest (Negate _) (Negate _)  = True
negTest (Negate _) e | isLit e = True
negTest e (Negate _) | isLit e = True
negTest _ _                    = False

-- The others are derived:

a /=* b = notE (a ==* b) 
a >*  b = b <* a
a <=* b = notE (b <* a)
a >=* b = b <=* a

(&&*), (||*) :: Exp -> Exp -> Exp

(&&*) = opt2C (&&#) (&&) And
 where
   e &&# LitBool False     = false
   LitBool False &&# e     = false
   e &&# LitBool True      = e
   LitBool True  &&# e     = e
   Not e &&# Not e'        = notE (e ||* e')
   e &&# e' | e ==% e'      = e
   e &&# e' | e ==% notE e' = false
   -- For the C back-end, which likes left-associative.
   e &&# (e' `And` e'')     = (e &&# e') &&# e''
   e &&# e'                = And e e'

(||*) = opt2C (||#) (||) Or
 where
   e ||# LitBool False     = e
   LitBool False ||# e     = e
   e ||# LitBool True      = true
   LitBool True  ||# e     = true
   Not e ||# Not e'        = notE (e &&* e')
   e ||# e' | e ==% e'      = e
   e ||# e' | e ==% notE e' = true
   -- For the C back-end, which likes left-associative.
   e ||# (e' `Or` e'')     = (e ||# e') ||# e''
   e ||# e'                = Or e e'

notE :: Exp -> Exp
notE = opt1_noIf notZ not
 where
   notZ (Not e) = e
   notZ e       = Not e



-----  Numbers

-- Bogus instances needed for Num, etc.
instance Enum Exp
instance Real Exp

type Unop  a = a -> a
type Binop a = a -> a -> a

instance Num Exp where
  (+) = -- (+%)
        opt2 (+#) (((+) :: Binop Int, (+) :: Binop Double), (+) :: Binop V4.F32vec4)
   where
     -- Move literals to right.  Since we've already done constant folding, 
     -- we won't make a loop here
     e@(LitInt    _) +# e'          = e' + e
     e@(LitDouble _) +# e'          = e' + e
     e@(PackF32 (LitDouble _) (LitDouble _) (LitDouble _) (LitDouble _)) +# e' = e' + e
     -- Combine literals.  (Can only appear as right argument.)
     (e `Add` e'@(LitInt    n)) +# e''@(LitInt    _) = e + (e' + e'')
     (e `Add` e'@(LitDouble _)) +# e''@(LitDouble _) = e + (e' + e'')
     (e `Add` e'@(PackF32 (LitDouble _) (LitDouble _) (LitDouble _) (LitDouble _)))
       +# e''@(PackF32 (LitDouble _) (LitDouble _) (LitDouble _) (LitDouble _)) = e + (e' + e'')
     -- Pull literals out of left summand
     (e `Add` e'@(LitInt    _)) +# e'' = (e + e'') + e'
     (e `Add` e'@(LitDouble _)) +# e'' = (e + e'') + e'
     (e `Add` e'@(PackF32 (LitDouble _) (LitDouble _)
                          (LitDouble _) (LitDouble _))) +# e'' = (e + e'') + e''
     -- Remove zeros
     e +# LitInt    0                  = e
     e +# LitDouble 0                  = e
     e +# PackF32 (LitDouble 0) (LitDouble 0) (LitDouble 0) (LitDouble 0) = e
     -- Cancellation
     e +# Negate e'       | e ==% e' = zero (typeOf e)
     Negate e +# e'       | e ==% e' = zero (typeOf e)
     -- Anti-distribution
     Mul a b +# Mul a' b' | a ==% a' = a * (b + b')
     Mul a b +# Mul a' b' | b ==% b' = (a + a') * b
     I2f a +# I2f b                 = i2f (a + b)
     I2f a +# LitDouble x | r == 0 = i2f (a + LitInt k)
       where (k,r) = properFraction x
     -- Very specialized: 1.0 - fromBF b = fromBF (255 - b)
--   LitDouble 1 +# Negate (Mul (I2f b) (LitDouble d)) | d == ov255
--                                  = fromBF (LitInt 255 - b)
     -- If nothing else works, try reassociating to the left.  Helps to
     -- expose constants for anti-distribution.  Also reduces parens when
     -- pretty-printed, since "+" is left-associative.
     e +# (e' `Add` e'')            = (e + e') + e''
     e +# e'                        = Add e e'

  (*) = opt2C (*#) (((*) :: Binop Int, (*) :: Binop Double), (*) :: Binop V4.F32vec4) Mul
   where
     -- Move literals to right.  Since we've already done constant folding, 
     -- we won't make a loop here
     e@(LitInt    _) *# e'          = e' * e
     e@(LitDouble _) *# e'          = e' * e
     e@(PackF32 (LitDouble _) (LitDouble _) (LitDouble _) (LitDouble _)) *# e' = e' * e

     -- Introduce exponents.  Do before re-association in case e is a product.
     Pow e n *# e' | e ==% e' = e ** (n + one Double)
     e' *# Pow e n | e ==% e' = e ** (n + one Double)
     Pow e n *# Pow e' n' | e ==% e' = e ** (n + n')
     e *# e'       | e ==% e' = e ** LitDouble 2

     -- Reassociate to the left.  Helps to expose constants for
     -- anti-distribution.  Also reduces parens when pretty-printed, since
     -- "*" is left-associative.
     e *# (e' `Mul` e'')            = (e * e') * e''
     -- Combine literals.  (Can only appear as right argument.)
     (e `Mul` e'@(LitInt    n)) *# e''@(LitInt    _) = e * (e' * e'')
     (e `Mul` e'@(LitDouble _)) *# e''@(LitDouble _) = e * (e' * e'')
     (e `Mul` e'@(PackF32 (LitDouble _) (LitDouble _) (LitDouble _) (LitDouble _))) *# e''@(PackF32 (LitDouble _) (LitDouble _) (LitDouble _) (LitDouble _)) = e * (e' * e'')
     -- Pull literals out of left summand
     (e `Mul` e'@(LitInt    _)) *# e'' = (e * e'') * e'
     (e `Mul` e'@(LitDouble _)) *# e'' = (e * e'') * e'
     (e `Mul` e'@(PackF32 (LitDouble _) (LitDouble _) (LitDouble _) (LitDouble _))) *# e'' = (e * e'') * e''
     -- Handle zeros and units
     e *# LitInt 1           = e
     e *# LitInt (-1)        = negate e
     e *# LitInt 0           = zero (typeOf e)
     e *# LitDouble 1        = e
     e *# LitDouble (-1)     = negate e
     e *# LitDouble 0        = zero (typeOf e)
     -- (conal) hmmm, why the typeOf uses above.  If we're type-correct,
     -- we should know the types.
     e *# PackF32 (LitDouble 0) (LitDouble 0) (LitDouble 0) (LitDouble 0) = f32vec4Zero
     e *# PackF32 (LitDouble 1) (LitDouble 1) (LitDouble 1) (LitDouble 1) = e

     -- 2^k * e ==> e <<< k.    Moved to compiler, so make expressions
     -- more regular during optimization.
     -- LitInt n *# m  | r == 0 = m <<< LitInt k
     --   where (k, r) = properFraction (logBase 2 (fromInt n))

     -- Restricted distributive rule, to combine literals.  We know x*e
     -- will get folded, so still one add and one multiply, but then we
     -- may get more simplification in x*e'.
     (e `Add` e'@(LitInt    _)) *# x@(LitInt    _) = e*x + e'*x
     (e `Add` e'@(LitDouble _)) *# x@(LitDouble _) = e*x + e'*x
     (e `Add` e'@(PackF32 (LitDouble _) (LitDouble _) (LitDouble _) (LitDouble _))) *# x@(PackF32 (LitDouble _) (LitDouble _) (LitDouble _) (LitDouble _)) = e*x + e'*x
     -- Inverses
     e *# Recip e' | e ==% e' = one  (typeOf e)
     Recip e *# e' | e ==% e' = one  (typeOf e)

     I2f e *# I2f e'         = i2f (e * e')

     e *# e'                 = Mul e e'

  negate = opt1C negateZ ((negate :: Unop Int, negate :: Unop Double),
                         negate :: Unop V4.F32vec4) Negate
   where
     negateZ (Negate e)   = e
     -- Push negate into right multiplicand, to help expose literals
     negateZ (e `Mul` e') = e * negate e'
     negateZ (e `Add` e') = negate e + negate e'
     -- Since i2f truncates toward zero ..
     negateZ (I2f e)      = i2f (- e)
     negateZ e            = Negate e

  abs = opt1C absZ (abs :: Unop Int, abs :: Unop Double) Abs
   where
     absZ e@(Abs _) = e
     absZ e         = Abs e

  fromInteger = error "No fromInteger for DExp.Exp"

l256 = LitInt 256

ov255 = recip 255 :: Double

-- Especially useful in conjunction with fromBF
nudgeToUnity :: Double -> Double
nudgeToUnity x | abs (x - 1) < 1.0e-6 = 1
               | otherwise            = x

instance Integral Exp where
  mod = opt2C modZ (mod :: Binop Int) Mod
          where
            -- e `mod` 2^k ==> e & (2^k-1).  Moved to compiler, so make
            -- expressions more regular during optimization.
            -- m `modZ` LitInt n | r == 0 = m .&. LitInt (n - 1)
            --   where (_, r) = properFraction (logBase 2 (fromInt n))
            m `modZ` n = Mod m n
  div = opt2C divZ (div :: Binop Int) Div
         where
           Mul n m `divZ` n' | n ==% n' = m
           Mul n m `divZ` m' | m ==% m' = n
            -- e `div` 2^k ==> e >>> k.  Moved to compiler, so make
            -- expressions more regular during optimization.
           -- m `divZ` LitInt n | r == 0 = m >>> LitInt k
           --  where (k, r) = properFraction (logBase 2 (fromInt n))
           n `divZ` m                  = Div n m

  -- Undefined:
  -- quotRem, divMod     :: a -> a -> (a,a)
  -- even, odd           :: a -> Bool
  -- toInteger           :: a -> Integer
  -- toInt               :: a -> Int


instance Fractional Exp where
  -- Use default for /:  a/b = a * recip b
  recip = opt1C recipZ (recip :: Unop Double, recip :: Unop V4.F32vec4) Recip
           where
             recipZ (Recip e)          = e
             recipZ (Mul e (Recip e')) = e'/e
             recipZ (Pow b c)          = Pow b (-c)
             recipZ e                  = Recip e
  fromRational = error "fromRational: no method for Exp"


-- floorE is defined in Exp.hs.


instance Floating Exp where
  pi   = LitDouble pi
  sin  = opt1C sinZ  (sin' :: Unop Double) Sin
          where sinZ (Negate e) = - sin e
                sinZ e          = Sin e
  cos  = opt1C cosZ  (cos' :: Unop Double) Cos
          where cosZ (Negate e) = cos e
                cosZ e          = Cos e
{-
  sqrt = opt1C Sqrt (sqrtSafe :: Unop Double, sqrt :: Unop V4.F32vec4) Sqrt
          where sqrtSafe x | x < 0     = trace ("oops: sqrt " ++ show x ++ "\n") $
                                         sqrt (-x)
                           | otherwise = sqrt x
  log  = opt1C logZ (log :: Unop Double) Log
          where logZ (Exp e) = e
                logZ e       = Log e
  exp  = opt1C Exp (exp :: Unop Double) Exp
-}
  logBase = opt2C logBaseZ (logBase :: Binop Double) LogBase
            where logBaseZ a b | a == b = one Double
                  logBaseZ b (Pow a c)  = logBaseZ b a * c
                  logBaseZ b (Mul a c)
                    | isLit b && (
                                     isLit a || isLit b  -- Worth transforming?
                                  || a == b  || c == b
                                 )
                    = logBaseZ b a + logBaseZ b c
                  logBaseZ b (Recip a)  = negate (logBaseZ b a)
                  logBaseZ b a          = LogBase b a
  (**) = opt2C (**#) ((**) :: Double -> Double -> Double) Pow
   where
     e **# LitDouble 0         = one Double
     e **# LitDouble 1         = e
     e **# LitDouble z | z < 0 = recip (e **# LitDouble (-z))
     Pow b a **# c             = b **# (a*c)
     e **# e'                  = e `Pow` e'

  -- sqrt uses default.
  log = logBase eDouble
  exp = (eDouble **)

  -- Add more later

eDouble :: Exp
eDouble = LitDouble (exp 1)


sin' = trig sin
cos' = trig cos

trig :: (Double -> Double) -> (Double -> Double)
trig f x = shave (f x)

-- shave off error
shave :: Double -> Double
shave x = 
  -- Multiply y by a power of two and see if the result is
  -- nearly an integer.  If so, scale back down integer value.  Power
  -- of two is so we get exact scaling down, and because values of 0, 1
  -- -1, and 0.5 are common as trig answers.
  let
      y  = x * s
      yR = fromInt (round y)
      x' = yR / s
      -- Hugs doesn't really have doubles.  Hmm... the error should really
      -- depend on the number's size.
      eps = 1.0e-6
      -- Scale factor.  If one, round to whole, if two, to half, etc.
      s   = 2
  in
      if abs (yR - y) < eps then
         x'
      else x

{-
trig :: (Double -> Double) -> Double -> (Double -> Double)
trig f atPi x | x == pi   = atPi
              | otherwise = f x
-}

instance RealFrac Exp

instance RealFloat Exp where
  atan2 = opt2C Atan2 (atan2 :: Binop Double) Atan2


---- Int/float conversions

i2f :: Exp -> Exp
i2f = opt1C I2f (fromInt :: Int -> Double) I2f

f2i :: Exp -> Exp
f2i = opt1C f2iZ (floor :: Double -> Int) F2i
 where
   f2iZ (I2f e)          = e
   -- The following two rules speed things up a lot at least in some
   -- cases.  Unfortunately, they're incorrect when e and e' have
   -- different signs.  Revisit when we switch from fi2 to round or floor
   -- as the primitive.
   -- f2iZ (e `Add` I2f e')     = f2i e + e'
   -- f2iZ (I2f e' `Add` e)     = f2i e + e'
   -- When 0=<e'<256, I think the is almost true (+-1).  Useful for lerpC.
   -- Not used with the F32vec4 color rep, but we may want to change back
   -- and use MMX.
   -- f2iZ (e `Mul` I2f e')     = (f2i (f256 * e) * e') `div` l256
   -- Oops :: "f2i (10 * i2f (var "i"))" loops infinitely when the next
   -- rule is in there.
   --    f2iZ (e@(I2f _) `Mul` e') = f2i (e' * e)
   f2iZ e                    = F2i e

   f256 = LitDouble 256

b2i :: Exp -> Exp
b2i = opt1C B2i (b2iN :: Bool -> Int) B2i
 where
   b2iN True  = 1
   b2iN False = 0



---- Bit-level operations.

-- Use Word32 instead of Int32 versions of <<< and >>> for
-- constant-folding, because we want logical not arithmetic shifts.

intify1 :: (Word32 -> Word32) -> (Int -> Int)
intify1 op a = toInt (op (fromInt a))

intify2 :: (Word32 -> Word32 -> Word32) -> (Int -> Int -> Int)
intify2 op a b = toInt (fromInt a `op` fromInt b)

instance Bits (Exp) where
  (.&.) = opt2C (.&.#) (intify2 (.&.)) BitAnd
   where
     b .&.# a@(LitInt 0) = a
     a@(LitInt 0) .&.# b = a
     b .&.# b'           = b `BitAnd` b'

  (.|.) = opt2C (.|.#) (intify2 (.|.)) BitOr
   where
     b .|.# LitInt 0 = b
     LitInt 0 .|.# b = b
     -- (m .&. n) .|. (m .&. n') = m .&. (n .|. n')
     (m `BitAnd` n) .|.# (m' `BitAnd` n') | m ==% m' = m .&. (n .|. n')
     (m `BitAnd` n) .|.# (m' `BitAnd` n') | n ==% n' = (m .|. m') .&. n
     -- right-associate
     -- (l `BitOr` m) .|.# n = l .|. (m .|. n)
     b .|.# b'       = BitOr b b'

  complement = opt1C complementZ (intify1 complement) BitNot
   where
     complementZ (BitNot b) = b
     complementZ b          = BitNot b

-- I'd like to add rules for &'ing and |'ing with 255 as well.
-- Questionable though, since the type says int, not UInt8.

-- Left shift
(<<<), (>>>) :: Exp -> Exp -> Exp

(<<<) = opt2C (<<<#) intShiftL LShift
 where
   bits <<<# LitInt 0   = bits
   m@(LitInt 0) <<<# n  = m
   LShift bits m <<<# n = bits <<< (n + m)
   {-
   -- Flattens results of (toRGB24 . fromRGB24)
   -- (m <<< k .|. n) <<< k'  ==>  m <<< (k+k') .|. n <<< k'
   ((m `LShift` k@(LitInt _)) `BitOr` n) <<<# k'@(LitInt _) =
     m <<< (k+k') .|. n <<< k'
   -}
   -- Do more generally, though might pessimize a bit
   (m `BitOr` n) <<<# k  = m <<< k .|. n <<< k
   -- (m >>> k .&. 255) <<< k  ==>  m && (255 <<< k) 
   -- ToDo: generalize.
   ((m `RShift` k@(LitInt _)) `BitAnd` l255@(LitInt 255)) <<<# k'
     | k == k' = m .&. (l255 <<< k) 
   bits <<<# n          = LShift bits n

   intShiftL bits n = toInt (shiftL (fromInt bits :: Word32) n)

-- Right shift
-- Right shift
--bits >>> n = bits <<< (-n)

(>>>) = opt2C (>>>#) intShiftR RShift
 where
   bits >>># LitInt 0   = bits
   m@(LitInt 0) >>># n  = m
   RShift bits m >>># n = bits >>> (n + m)
   bits >>># n          = RShift bits n

   intShiftR bits n = toInt (shiftR (fromInt bits :: Word32) n)

-- F32vec4

packF32 = opt4C PackF32 V4.packF32 PackF32

nthF32  = opt2C nthF32Z V4.nthF32 NthF32
 where
   nthF32Z (LitInt i) _ | i < 0 || i > 3 = error "nthF32: index out of range"
   nthF32Z (LitInt i) (PackF32 a b c d) = [a,b,c,d]!!i
   nthF32Z i c = NthF32 i c

sumF32 = opt1C SumF32 V4.sumF32 SumF32

-- Tuples

mkTuple :: [Exp] -> Exp
mkTuple = MkTuple

-- MkTuple should probably do mkTuple [tupleElt 0 e, ..., tupleElt (n-1) e]
-- where e is an n-tuple.  We don't have opt functions for argument lists,
-- but we could.

tupleElt :: Int -> Exp -> Exp
tupleElt i = opt1C tupleEltZ NoFold (TupleElt i)
 where
   tupleEltZ (MkTuple es) = es!!i
   tupleEltZ e            = TupleElt i e

-- Misc

deref8, deref32 :: Exp -> Exp
-- Float conditionals, but don't do constant folding

deref8  = opt1C Deref8  NoFold Deref8
deref32 = opt1C Deref32 NoFold Deref32

-- deref8  (If c a b)  = ifE c (deref8  a) (deref8  b)
-- deref8  addr        = Deref8  addr

-- deref32 (If c a b)  = ifE c (deref32 a) (deref32 b)
-- deref32 addr        = Deref32 addr

-- deref8  = Deref8
-- deref32 = Deref32

-- Construct a new array, given the size (but *not* an
-- initial value.)
newArr :: Exp -> Type -> Exp
newArr sz ty = NewArr sz ty

-- Make an array, initializing the members.
mkArr :: Exp -> (Exp -> Exp) -> Exp
mkArr size elem = MkArr size i (elem (Var i Int))
 where
   i  = unsafePerformIO (newId' "i")

   -- Smart constructor after lambda elim.  Not used above, because we'd
   -- need to verify that arr has size n.
   mkArrZ n i (ReadArr arr (Var j _))
     | i == j && not (occFreeIn i arr) = arr
   mkArrZ n i e = MkArr n i e

-- We cannot determine the type, so require it given
lambda :: Type -> (Exp -> Exp) -> Exp
lambda ty body = Lambda x ty (body (Var x ty))
 where
   x = unsafePerformIO (newId' "x")


-- We cannot determine the type, so require it given
lambdaVar :: Id -> Type -> (Exp -> Exp) -> Exp
lambdaVar x ty body = Lambda x ty (body (Var x ty))


---- Imperative stuff

newRef :: Exp -> Exp -- (Exp -> Exp) -> Exp
newRef init = NewRef init -- (lambda (Ref (typeOf init)) body)

readRef :: Exp -> Exp
readRef ref = ReadRef ref

writeRef :: Exp -> Exp -> Exp
writeRef ref newVal = WriteRef ref newVal

{-
updateRef :: Exp -> (Exp -> Exp) -> Exp
updateRef ref updater = UpdateRef ref (lambda ty updater)
 where
   Ref ty = typeOf ref
-}

setInt :: Exp -> Exp -> Exp
setInt = opt2C SetInt NoFold SetInt

-- Or if we don't even want if-floating:
-- setInt = SetInt

noop :: Exp
noop = Noop

{-
(>>*) :: Exp -> (Exp -> Exp) -> Exp
(>>*) a b = (>>=*) a (Lambda  
opt2C (>>#) NoFold Then
 where
   Noop >># a = a
   a >># Noop = a
   a >># b    = Bind a (const b)
-}

(>>=*) :: Exp -> Exp -> Exp
(>>=*) = opt2C (>>=#) NoFold Bind
 where
   Noop >>=# (Lambda _ _ a) = a
   a >>=# (Lambda _ _ Noop) = a
   a >>=# b  = Bind a b

-- Given loop limit n and body f, do f 0; ...; f (n-1)
loop :: Exp -> (Exp -> Exp) -> Exp

loop limit body = loop' (body (Var itvar Int))
 where
   -- Float ifs above loop when legal, i.e., the condition doesn't use the
   -- loop variable.  (Should do something similar for mkArr.)
   loop' (If c a b) | noItvar c = If c (loop' a) (loop' b)
                    | isJust mb = fromJust mb
    where
      mb = tryCopyPatch itvar c a b

   loop' exp = Loop itvar limit exp
   itvar     = unsafePerformIO (newId' "k")
   noItvar e = not (occFreeIn itvar e)


doProgress :: Exp -> Exp -> Exp
doProgress v w = DoProgress v w

copy1DArray :: Exp -> Exp -> Exp
copy1DArray v w = Copy1DArray v w

tryCopyPatch :: Id -> Exp -> Exp -> Exp -> Maybe Exp
tryCopyPatch itvar c a b = Nothing

{- working here

tryCopyPatch itvar c a b = Nothing

        | j' == j && noItvar a  =

((j `LessThan` lo) `Or` (hi `LessThan` j'))     
-}

--    loop' (If c (WriteArr arr i val) (WriteArr arr' i' val'))
--       | arr == arr' && i == i' && 

-----------  Experimental approach 

-- N-ary operations on expressions
type EOp1 = Exp -> Exp
type EOp2 = Exp -> Exp -> Exp
type EOp3 = Exp -> Exp -> Exp -> Exp
type EOp4 = Exp -> Exp -> Exp -> Exp -> Exp

-- N-ary possibly optimizing operations
type Opt1 = Exp -> Maybe Exp
type Opt2 = Exp -> Exp -> Maybe Exp
type Opt3 = Exp -> Exp -> Exp -> Maybe Exp
type Opt4 = Exp -> Exp -> Exp -> Exp -> Maybe Exp

class Opt opt where thenTry :: opt -> opt -> opt

instance Opt Opt1 where
  thenTry optA optB = \ e1 -> optA e1 `mplus` optB e1

instance Opt Opt2 where
  thenTry optA optB = \ e1 e2 -> optA e1 e2 `mplus` optB e1 e2

instance Opt Opt3 where
  thenTry optA optB = \ e1 e2 e3 -> optA e1 e2 e3 `mplus` optB e1 e2 e3

instance Opt Opt4 where
  thenTry optA optB = \ e1 e2 e3 e4 -> optA e1 e2 e3 e4 `mplus` optB e1 e2 e3 e4


-- Do all if-floating, then try an optimizer, using given constructor when
-- the optimizer fails.

floatIf1 :: Opt1 -> EOp1 -> EOp1
floatIf1 opt con = h
 where
  h (If c a b) = ifE c (h a) (h b)
  h e1         = maybe (con e1) id (opt e1)

floatIf2 :: Opt2 -> EOp2 -> EOp2
floatIf2 opt con = h
 where
  h (If c a b) e2 | okFloatIf [e2] = ifE c (h a e2) (h b e2)
  h e1 (If c a b) | okFloatIf [e1] = ifE c (h e1 a) (h e1 b)
  h e1 e2 = maybe (con e1 e2) id (opt e1 e2)

floatIf3 :: Opt3 -> EOp3 -> EOp3
floatIf3 opt con = h
 where
  h (If c a b) e2 e3 | okFloatIf [e2,e3] = ifE c (h a e2 e3) (h b e2 e3)
  h e1 (If c a b) e3 | okFloatIf [e1,e3] = ifE c (h e1 a e3) (h e1 b e3)
  h e1 e2 (If c a b) | okFloatIf [e1,e2] = ifE c (h e1 e2 a) (h e1 e2 b)
  h e1 e2 e3 = maybe (con e1 e2 e3) id (opt e1 e2 e3)

floatIf4 :: Opt4 -> EOp4 -> EOp4
floatIf4 opt con = h
 where
  h (If c a b) e2 e3 e4 | okFloatIf [e2,e3,e4] = ifE c (h a e2 e3 e4) (h b e2 e3 e4)
  h e1 (If c a b) e3 e4 | okFloatIf [e1,e3,e4] = ifE c (h e1 a e3 e4) (h e1 b e3 e4)
  h e1 e2 (If c a b) e4 | okFloatIf [e1,e2,e4] = ifE c (h e1 e2 a e4) (h e1 e2 b e4)
  h e1 e2 e3 (If c a b) | okFloatIf [e1,e2,e3] = ifE c (h e1 e2 e3 a) (h e1 e2 e3 b)
  h e1 e2 e3 e4 = maybe (con e1 e2 e3 e4) id (opt e1 e2 e3 e4)


-- Given constant folder and other optimizer, if we're optimizing, try
-- constant folding and optimizing, otherwise just folding.
tryOpt :: Opt opt => opt -> opt -> opt
tryOpt | optimize  = \ cfold opt -> cfold `thenTry` opt
       | otherwise = \ cfold opt -> cfold

(+%) :: EOp2

(+%) = floatIf2 (+#) Add

(+#) = cfold2 (((+) :: Binop Int, (+) :: Binop Double)
               ,(+) :: Binop V4.F32vec4)
        `tryOpt` (+##)

-- Here's the strategy for e + e': If right-associated,
-- left-associate.  If neither addend is a sum, try to combine them
-- directly.  If either is a sum, say e + (e' + e''), try (e+e')+e''
-- or (e+e'')+e'.

-- Left-associate (fewer parens in unparsed and fewer registers in
-- evaluation).  [Somewhat wasteful to use "+" here, which will again
-- consider if-floating.]
e +## (e' `Add` e'') = Just ((e + e') + e'')

-- A/C rearrangement.  Try (e'+e'')+e or (e+e'')+e'
(e `Add` e') +## e'' = fmap (+ e ) (e' +# e'') `mplus`
                       fmap (+ e') (e  +# e'')

----   Direct cases:

-- Zero/inverse
e +## e' 
  | isZero e  = Just e'
  | isZero e' = Just e
  | e ==% -e' = Just $ zero (typeOf e) -- (w/ smart negate)

-- Commute literal to left, to reduce cases below
-- e +## lit | isLit e' = lit +## e

-- Anti-distribution
Mul a b +## Mul a' b' | a ==%  a' = Just $ a * (b + b')
                      | a ==% -a' = Just $ a * (b - b')
                      | b ==%  b' = Just $ (a + a') * b
                      | b ==% -b' = Just $ (a - a') * b
-- Swap float op for int op.  Having i2f at the top provides more
-- manifest info for further simplification.
I2f a +## I2f b                 = Just $ i2f (a + b)
I2f a +## LitDouble x | r == 0  = Just $ i2f (a + LitInt k)
  where (k,r) = properFraction x
LitDouble x +## I2f a | r == 0  = Just $ i2f (LitInt k + a)
  where (k,r) = properFraction x

-- Otherwise, fail
_ +## _ = Nothing


isZero (LitInt    0) = True
isZero (LitDouble 0) = True
isZero (PackF32 (LitDouble 0) (LitDouble 0) (LitDouble 0) (LitDouble 0)) = True
isZero _ = False


-- Like DExp.fromHnf, but use smart constructors instead of dumb ones.

fromHnfSmart :: (Op, [Exp]) -> Exp

fromHnfSmart (AddOp, [e,e'])               = e + e'
fromHnfSmart (NegateOp, [e])               = - e
fromHnfSmart (SubOp, [e,e'])               = e - e'
fromHnfSmart (MulOp, [e,e'])               = e * e'
fromHnfSmart (RecipOp, [e])                = recip e
fromHnfSmart (DivideOp, [e,e'])            = e / e'
fromHnfSmart (PowOp, [e,e'])               = e ** e'
fromHnfSmart (LogBaseOp, [e,e'])           = logBase e e'
fromHnfSmart (SinOp, [e])                  = sin e
fromHnfSmart (Atan2Op, [e,e'])             = atan2 e e'
fromHnfSmart (CosOp, [e])                  = cos e
fromHnfSmart (I2fOp, [e])                  = i2f e
fromHnfSmart (F2iOp, [e])                  = f2i e
fromHnfSmart (B2iOp, [e])                  = b2i e
fromHnfSmart (ModOp, [e,e'])               = mod e e'
fromHnfSmart (DivOp, [e,e'])               = e `div` e'
fromHnfSmart (EqualOp, [e,e'])             = e ==* e'
fromHnfSmart (LessThanOp, [e,e'])          = e <* e'
fromHnfSmart (MinOp, [e,e'])               = min e e'
fromHnfSmart (MaxOp, [e,e'])               = max e e'
fromHnfSmart (AndOp, [e,e'])               = e &&* e'
fromHnfSmart (OrOp, [e,e'])                = e ||* e'
fromHnfSmart (NotOp, [e])                  = notE e
fromHnfSmart (LShiftOp, [e,e'])            = e <<< e'
fromHnfSmart (RShiftOp, [e,e'])            = e >>> e'
fromHnfSmart (BitAndOp, [e,e'])            = e .&. e'
fromHnfSmart (BitOrOp, [e,e'])             = e .|. e'
fromHnfSmart (BitNotOp, [e])               = complement e
fromHnfSmart (PackF32Op, [e1, e2, e3, e4]) = packF32 e1 e2 e3 e4
fromHnfSmart (NthF32Op, [e,e'])            = nthF32 e e'
fromHnfSmart (MkTupleOp, es)               = mkTuple es
fromHnfSmart (TupleEltOp n, [e])           = tupleElt n e
fromHnfSmart (SumF32Op, [e])               = sumF32 e
fromHnfSmart (SetIntOp, [e,e'])            = setInt e e'
fromHnfSmart (Deref8Op, [e])               = deref8 e
fromHnfSmart (Deref32Op, [e])              = deref32 e
fromHnfSmart (NoopOp,[])                   = noop
fromHnfSmart (BindOp, [e,e'])              = e >>=* e'
fromHnfSmart (ReadArrOp _, [e,e'])         = ReadArr e e'  -- why dumb? (conal)
fromHnfSmart (NewRefOp, [e])               = NewRef e
fromHnfSmart (ReadRefOp, [e])              = ReadRef e
fromHnfSmart (WriteRefOp, [e,e'])          = WriteRef e e'
fromHnfSmart (IfOp, [e,e',e''])            = ifE e e' e''
fromHnfSmart (DoProgressOp, [a,b])         = doProgress a b
fromHnfSmart (Copy1DArrayOp, [a,b])        = copy1DArray a b

