---- Type checking.  For finding compiler bugs.

module TCheck where

import Monad
import IOExts (trace)
import Types
import DExp
import PP
import PPDExp (showsPrecDExp)


-- Error monad
data ErrM a = Fail String | Ok a

instance Monad ErrM where
  return         = Ok
  Ok   a   >>= f = f a
  Fail msg >>= _ = Fail msg
  fail           = Fail

instance Functor ErrM where
  -- Standard definition.
  fmap f m = m >>= return . f

runEM            :: String -> ErrM a -> a
runEM _   (Ok a)     = a
runEM str (Fail msg) = error $ str ++ "\n" ++ msg

class TCheck a where
  -- Returns given exp if type-correct, and error otherwise.
  -- For debugging Pan.  The given string can name a compiler pass, etc.
  tcheck :: String -> a -> a

-- More convenient, but less informative.
tck e = tcheck "Check" e

instance TCheck b => TCheck (a -> b) where
  tcheck str f = \ x -> tcheck str (f x)

instance (TCheck a, TCheck b) => TCheck (a,b) where
  tcheck str (a,b) = (tcheck str a, tcheck str b)

instance (TCheck a, TCheck b, TCheck c) => TCheck (a,b,c) where
  tcheck str (a,b,c) = (tcheck str a, tcheck str b, tcheck str c)

instance (TCheck a, TCheck b, TCheck c, TCheck d) => TCheck (a,b,c,d) where
  tcheck str (a,b,c,d) =
    (tcheck str a, tcheck str b ,tcheck str c, tcheck str d)



instance TCheck Exp where
  tcheck str e = runEM str (typeCheck e) `seq` e



requireEM :: Bool -> String -> ErrM ()
requireEM True  _   = Ok ()
requireEM False msg = fail msg

-- ToDo: replace "show e" with pretty printed

errmsg :: Exp -> String -> String
errmsg e msg = msg ++ " for exp:\n" ++ show e

sameCheck :: Exp -> Type -> Type -> ErrM ()
sameCheck a aty bty = requireEM (aty == bty) wrongTy
 where
   wrongTy = errmsg a $ "Type-mismatch: got " ++ show aty
                     ++ " but expected "      ++ show bty


-- Top-level checks for array, ref or action, using just typeOf.  Must be
-- followed by deep checks.  Yields type parameter
arrCheck, refCheck, actCheck :: Exp -> ErrM Type
funCheck                     :: Exp -> ErrM (Type,Type)

arrCheck e = do requireEM ok (errmsg e "Non-array")
                return valTy
 where
   (ok,valTy) = case typeOf e of
                Arr ty -> (True,ty)
                _      -> (False,undefined)

refCheck e = do requireEM ok (errmsg e "Non-reference")
                return valTy
 where
   (ok,valTy) = case typeOf e of
                Ref ty -> (True,ty)
                _      -> (False,undefined)

actCheck e = do requireEM ok (errmsg e "Non-action")
                return valTy
 where
   (ok,valTy) = case typeOf e of
                Action ty -> (True,ty)
                _         -> (False,undefined)

funCheck e = do requireEM ok (errmsg e "Non-function")
                return (dom,ran)
 where
   (ok,dom,ran) = case typeOf e of
                  Fun dom ran -> (True,dom,ran)
                  _           -> (False,undefined,undefined)


-- Similar checkClass :: (Type -> Bool) -> String -> Exp -> ErrM Type
checkClass pred expected a = do requireEM (pred ty) msg
                                return ty
 where
   ty  = typeOf a
   msg = errmsg a ("Expected " ++ expected)
                           

isDoubleTy Double = True
isDoubleTy _      = False

isFloatingTy F32vec4 = True
isFloatingTy ty      = isDoubleTy ty

isNumTy Int = True
isNumTy ty  = isFloatingTy ty



unOpTC :: (Type -> Bool) -> String -> Exp -> ErrM Type
unOpTC pred msg a = do ty <- checkClass pred msg a
                       sigTC [ty] ty [a]

-- Unary numeric operators.
unNumOpTC,unFloatingOpTC,unDoubleOpTC :: Exp -> ErrM Type
unNumOpTC      = unOpTC isNumTy      "number"
unFloatingOpTC = unOpTC isFloatingTy "floating"

-- Binary
binOpTC :: (Type -> Bool) -> String -> Exp -> Exp -> ErrM Type
binOpTC pred msg a b = do ty <- checkClass pred msg a
                          sigTC [ty,ty] ty [a,b]

-- Binary numeric operators.
binNumOpTC      = binOpTC isNumTy      "number"
binFloatingOpTC = binOpTC isFloatingTy "floating"

-- Num a => a -> Int -> a
numIntOpTC a b = do aTy <- checkClass isNumTy "number" a
                    sigTC [aTy,Int] aTy [a,b]

typeCheckAs :: Type -> Exp -> ErrM ()
typeCheckAs goalTy e = do ty <- typeCheck e
                          sameCheck e ty goalTy

-- Type-check using signature. Takes result type, argument types,
-- arguments.
sigTC :: [Type] -> Type -> [Exp] -> ErrM Type
sigTC [] resTy [] = return resTy
sigTC (ty:tys) resTy (e:es) = do typeCheckAs ty e
                                 sigTC tys resTy es


unIntOpTC    e = sigTC [Int   ]    Int [e]
unDoubleOpTC e = sigTC [Double] Double [e]
unBoolOpTC   e = sigTC [Bool  ]   Bool [e]

binIntOpTC    e e' = sigTC [Int   ,Int   ]    Int [e,e']
binDoubleOpTC e e' = sigTC [Double,Double] Double [e,e']
binBoolOpTC   e e' = sigTC [Bool  ,Bool  ]   Bool [e,e']

binRelOpTC e e' = sigTC [ty,ty] Bool [e,e']
                  where ty = typeOf e


typeCheck :: Exp -> ErrM Type

typeCheck (LitInt _)         = return Int
typeCheck (LitDouble _)      = return Double
typeCheck (LitBool _)        = return Bool
typeCheck (ImportData _)     = return Int
typeCheck (Var _ ty)         = return ty
typeCheck (Lambda x ty e)    = fmap (Fun ty) (typeCheck e)                                  
typeCheck (Let _ rhs e)      = do typeCheck rhs
                                  typeCheck e
typeCheck (If c a b)         = sigTC [Bool,ty,ty] ty [c,a,b]
                               where ty = typeOf a
typeCheck (Add e e')         = binNumOpTC e e'
typeCheck (Negate e)         = unNumOpTC e
typeCheck (Mul e e')         = binNumOpTC e e'
typeCheck (Recip e)          = unFloatingOpTC e
typeCheck (Pow e e')         = binDoubleOpTC e e'
typeCheck (LogBase e e')     = binDoubleOpTC e e'
typeCheck (Sin e)            = unDoubleOpTC e
typeCheck (Cos e)            = unDoubleOpTC e
typeCheck (Atan2 e e')       = binDoubleOpTC e e'
typeCheck (I2f e)            = sigTC [Int] Double [e]
typeCheck (F2i e)            = sigTC [Double] Int [e]
typeCheck (B2i e)            = sigTC [Bool] Int [e]
typeCheck (Abs e)            = unFloatingOpTC e
typeCheck (Mod e e')         = binIntOpTC e e'
typeCheck (Div e e')         = binIntOpTC e e'
typeCheck (Equal e e')       = binRelOpTC e e'
typeCheck (LessThan e e')    = binRelOpTC e e'
typeCheck (Min e e')         = binNumOpTC e e'
typeCheck (Max e e')         = binNumOpTC e e'
typeCheck (And e e')         = binBoolOpTC e e'
typeCheck (Or e e')          = binBoolOpTC e e'
typeCheck (Not e)            = unBoolOpTC e
typeCheck (LShift e e')      = binIntOpTC e e'
typeCheck (RShift e e')      = binIntOpTC e e'
typeCheck (BitAnd e e')      = binIntOpTC e e'
typeCheck (BitOr e e')       = binIntOpTC e e'
typeCheck (BitNot e)         = unIntOpTC e
typeCheck (PackF32 a b c d)  = sigTC [Double,Double,Double,Double] F32vec4 [a,b,c,d]
typeCheck (NthF32 e e')      = sigTC [F32vec4,Int] Double [e,e']
typeCheck (SumF32 e)         = sigTC [F32vec4] Double [e]
typeCheck (MkTuple es)       = liftM Tuple (mapM typeCheck es)
typeCheck (Deref8 e)         = unIntOpTC e
typeCheck (Deref32 e)        = unIntOpTC e
typeCheck (SetInt e e')      = sigTC [Int,Int] (Action Void) [e,e']
typeCheck b@(Bind e e')      = --trace ("typeCheck bind\n") $
                               -- bind :: Action a -> (a -> Action b) -> Action b
                               do a        <- actCheck e
                                  (_,actb) <- funCheck e'
                                  let (ok,b) = case actb of
                                               Action b -> (True,b)
                                               _        -> (False,undefined)
                                  requireEM ok
                                    (errmsg e' "Non-action-producing function")
                                  sigTC [Action a, Fun a (Action b)]
                                        (Action b) [e,e']
                               --return (Action Void) -- sigTC [Action ,Action] Action [e,e']
typeCheck (DoProgress a b)   = sigTC [Int,Int] (Action Void) [a,b]
typeCheck (Copy1DArray a b)  = sigTC [Int,Int] (Action Void) [a,b]
typeCheck Noop               = return (Action Void)
typeCheck (Loop _ n b)       = sigTC [Int,Action Void] (Action Void) [n,b]
typeCheck (ReadArr arr i)    = do valTy <- arrCheck arr
                                  sigTC [Arr valTy, Int] valTy [arr,i]
typeCheck (ReadMutArr arr i) = do valTy <- arrCheck arr
                                  sigTC [Arr valTy, Int] (Action valTy) [arr,i]
typeCheck (WriteArr arr i x) = do valTy <- arrCheck arr
                                  sigTC [Arr valTy, Int, valTy] (Action Void) [arr,i,x]
typeCheck e@(NewArr sz ty)   = --trace ("TCheck NewArr: " ++ show e) $
                               sigTC [Int] (Action (Arr ty)) [sz]
typeCheck (MkArr n _ elem)   = sigTC [Int,valTy] (Arr valTy) [n,elem]
                               where valTy = typeOf elem
typeCheck Rand               = return (Action Int)			       
typeCheck (NewRef init)      = sigTC [ty] (Action (Ref ty)) [init]
                               where ty = typeOf init
typeCheck (ReadRef r)        = do ty <- refCheck r
                                  sigTC [Ref ty] (Action ty) [r]
typeCheck (WriteRef r v)     = do ty <- refCheck r
                                  sigTC [Ref ty,ty] (Action Void) [r,v]
typeCheck (Perform act)      = do valTy <- actCheck act
                                  sigTC [Action valTy] valTy [act]
typeCheck (Return e)         = sigTC [a] (Action a) [e]
                               where a = typeOf e

typeCheck e = error ("TCheck.typeCheck: no case for " ++ show e)
